/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/JavaSpand/spand/Msg.java,v 1.11 1998/06/27 01:41:19 stemm Exp $ */
package spand;
import java.net.*;
import java.io.*;
import java.util.*;

/** A Spand message consists of a marshalled data object and a
 * handler. The handler either indicates the type of the marshalled
 * object (if active is false) or the name of the handler of the marshalled
 * data (if active is true).
 * @version $Revision: 1.11 $
 * @author SPAND Group, UC Berkeley.  */
public class Msg implements SendableSpandObject {

   /** The name of the handler. Stored as a fixed string for simplicity */
  public String handler;
  
  /** Whether the handler string indicates the handler or indicates
      the type of the payload */
  boolean active;

  /** The length of the marshalled object */
  int dataLen;

  /** The marshalled object data */
  byte[] data;

  /** The port that the message was sent from. Not actually included
     in the message (not marshalled/unmarshalled). */
  public int sendPort;
  
  /** The address that the message was sent from. Not actually
     included in the message (not marshalled/unmarshalled). */
  public InetAddress sendAddr;

  /** Default Constuctor */
  public Msg(){
  }

  /** Constructor from socket. Reads data from socket and unmarshalls Msg object */
  public Msg(DatagramSocket sock) throws IOException{
    DatagramPacket p = new DatagramPacket(new byte[Spand.MaxMsgSize], Spand.MaxMsgSize);
    sock.receive(p);
    sendPort=p.getPort();
    sendAddr=p.getAddress();
    Unmarshal(p.getData(), 0);
  }

  /** Constructor from Stream Socket */
  public Msg(Socket sock) throws IOException{
    sendAddr=sock.getInetAddress();
    sendPort=sock.getPort();
    InputStream in = sock.getInputStream();
    Unmarshal(in);
  }    
  
 /** Constructor from handler and payload */
  public Msg(String handler0, SendableSpandObject o, boolean active0){
    handler=handler0;
    data = new byte[o.Marshal(null, 0)];
    dataLen=o.Marshal(data, 0);
    active=active0;
  }
  
  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  public int Marshal(byte[] wireData, int startIndex){
    if (wireData==null){
      int unalign_size=12*dataLen+handler.length();
      /* 12 bytes for handler string length, data length, active flag */
      while((unalign_size%4)!=0){
	unalign_size++;
      }
      return (unalign_size);
    }
    
    int pos=startIndex;

    ArrayConv.StringToByteArray(wireData, pos, handler);
    pos+=(handler.length()+4);
    while((pos%4)!=0){
      pos++;
    }

    /* Write the Active Flag */
    ArrayConv.BooleanToByteArray(wireData, pos, active);
    pos+=4;
    
    /* Write the length of the data */
    ArrayConv.IntToByteArray(wireData, pos, dataLen);
    pos+=4;

    /* Now copy the message data */
    for (int i=0,j=pos; i<dataLen; i++, j++, pos++){
      wireData[j]=data[i];
    }

    return (pos-startIndex);
  }

  /** Unpackage this object from a stream of bytes. The number of
      bytes read is returned. */
  public int Unmarshal(InputStream in) throws IOException{
    byte sizeArr[] = new byte[4];
    byte handlerArr[];
    int size=0;
    int handlerLen;
    int i;

    /* Beginning of message is always size of handler, read that first */
    in.read(sizeArr, 0, 4);

    handlerLen=ArrayConv.ByteArrayToInt(sizeArr, 0);

    /* Copy size to handlerArr (so we can use ArrayConv to convert the string) */
    handlerArr = new byte[4+handlerLen];

    for(i=0; i<4; i++){
      handlerArr[i]=sizeArr[i];
    }

    /* Read handler string */
    in.read(handlerArr, 4, handlerLen);
    handler=ArrayConv.ByteArrayToString(handlerArr, 0);
    size+=(4+handlerLen);
    while((size%4)!=0){
      in.read();
      size++;
    }
    
    /* Read the Active Flag */
    in.read(sizeArr, 0, 4);
    active=ArrayConv.ByteArrayToBoolean(sizeArr, 0);
    size+=4;
    
    /* Read Data Length */
    in.read(sizeArr, 0, 4);
    dataLen=ArrayConv.ByteArrayToInt(sizeArr, 0);
    
    /* Read Data */
    data = new byte[dataLen];
    in.read(data, 0, dataLen);

    size+=(4+dataLen);
    
    return size;
  }

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  public int Unmarshal(byte[] wireData, int startIndex){

    int pos=startIndex;

    handler=ArrayConv.ByteArrayToString(wireData, pos);
    pos+=(handler.length()+4);
    while((pos%4)!=0){
      pos++;
    }
 
    /* Read the Active Flag */
    active=ArrayConv.ByteArrayToBoolean(wireData, pos);
    pos+=4;
    
    /* Read the length of the data */
    dataLen = ArrayConv.ByteArrayToInt(wireData, pos);
    pos+=4;

    data = new byte[dataLen];
    for (int i=0, j=pos; i<dataLen; i++, j++, pos++){
      data[i]=wireData[j];
    }
    
    return (pos-startIndex);
    
  }

  /** Turn this message into an object. */
  public Object toObject(){
    Object o = null;
    try {
      o = (Object) (Class.forName(this.handler)).newInstance();
    } catch (ClassNotFoundException e1) {
      System.err.println("Couldn't find class "+this.handler);
      return(null);
    } catch (InstantiationException e2) {
      System.err.println("Couldn't create class "+this.handler);
      return(null);
    } catch (IllegalAccessException e3) {
      System.err.println("Don't have permission to create class "+this.handler);
      return(null);
    }

    if (!active){
      /* The object o is a SendableSpandObject to execute */
      SendableSpandObject s = (SendableSpandObject) o;
      s.Unmarshal(data, 0);
      return s;
    } else {
      return o;
    }
  }
  
  /** Print out this message as a string */
  public String toString(){
    if (active){
      SpandHandler h = (SpandHandler) this.toObject();
      return handler+" "+h.toString(this);
    } else {
      return handler+" "+this.toObject().toString();
    }
  }

}


