/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/JavaSpand/spand/MsgHandler.java,v 1.3 1998/06/26 16:29:17 stemm Exp $ */
package spand;
import java.net.*;
import java.io.*;
import java.util.*;

/**
 * A MsgHandler is a thread that listens for SPAND Messages from
 * Clients. When created, the Handler specified in
 * the message is invoked, and if the Handler generates a reply
 * message, the reply is returned to the sender. 
 * @version $Revision: 1.3 $
 * @author SPAND Group, UC Berkeley.  */

public class MsgHandler extends Thread{

    /** Used to listen for Datagram Requests */
    DatagramSocket requestDgramSock;
  
    /** Holds the active Stream Connection */
    Socket activeConn;

    /** Holds msg to be handled */
    Msg msg;

    /** The database of Performance Reports. The hash maps from String
      objects (which indicate the type of Performance Report) to
      arbitrary Java Objects. The SpandHandlers that use a particular
      Performance Report will know what the type of the value in the
      hash is. */
    Hashtable globalHash;

    /** If true, then use a stream socket to listen for SPAND Messages. */
    final boolean stream; 

    /** File For logging messages */
    FileWriter logFile;
  
    /** Create the listener. globalHash0 is an already-created
      Hashtable, and stream0 indicates whether to listen on a Stream
      or Datagram socket. */
    public MsgHandler(Msg msg0, Hashtable globalHash0, boolean stream0, 
		      DatagramSocket requestDgramSock0, 
		      Socket activeConn0, FileWriter logFile0) throws SocketException, IOException {

			  globalHash=globalHash0;
			  stream=stream0;
			  msg = msg0;
			  logFile=logFile0;

			  if (stream){
			      activeConn = activeConn0;
			  } else {
			      requestDgramSock = requestDgramSock0;
			  }
			  this.start();
    }

    /** Start the listener, receiving messages, running their handlers,
      and returning their reply messages. */
      
    public void run() {
	Msg reply = null;
	SpandHandler handler;
	
	/* Try to create the handler that they specified */
	handler = (SpandHandler) msg.toObject();
	reply=handler.execute(msg, globalHash);

	/* Print out the request and reply message */
	try {
	  if (reply==null){
	    logFile.write("*** ["+new Date().toString()+"] "+msg.sendAddr.toString()+"."+msg.sendPort+" "+msg.toString()+" <none>\n");
	    logFile.flush();
	  } else {
	    Date now = new Date();
	    logFile.write("*** ["+new Date().toString()+"] "+msg.sendAddr.toString()+"."+msg.sendPort+" "+msg.toString()+" "+reply.toString()+"\n");
	    logFile.flush();
	  }
	} catch (IOException e1){
	  System.err.println("Couldn't write to log file");
	}

	if (reply!=null){
	    SendReply(reply, msg.sendAddr, msg.sendPort);
	}
    }

    /** Send a reply message over either the stream or datagram socket. */
    private void SendReply(Msg reply, InetAddress replyAddr, int replyPort){

	byte[] buf;
	int size;
	DatagramPacket p;
    
	/* If we are using UDP as the transport and the size of the
	   reply won't fit in a single packet, reply with
	   "Reply too large for this transport protocol" error. */
	if (reply.Marshal(null, 0) >= Spand.MaxMsgSize && !stream){
	    /* Reply with "Reply too large" error */
	}
	
	buf = new byte[reply.Marshal(null, 0)*2];
	size=reply.Marshal(buf, 0);
	try {
	    if (stream){
		OutputStream out = activeConn.getOutputStream();
		out.write(buf);
		activeConn.close();
	    } else {
		p = new DatagramPacket(buf, size, replyAddr, replyPort);
		requestDgramSock.send(p);
	    }
	} catch (IOException e){
	  try {
	    logFile.write("+++ ["+new Date().toString()+"] "+msg.sendAddr.toString()+"."+msg.sendPort+" Couldn't send reply "+e.toString());
	    logFile.flush();
	  } catch (IOException e1){
	    System.err.println("Couldn't send reply to msg..continuing");
	  }
	}
    }
}




