/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/Socket++/UdpReliable.h,v 1.3 1998/06/08 15:14:56 ss Exp $ */
#ifndef SS_SOCKET_UDP_RELIABLE
#define SS_SOCKET_UDP_RELIABLE
#include "Socket.h"

#define SOCK_MAXWIND 32
#define SOCK_FULLMASK 0x00000001 << (SOCK_MAXWIND - 1)
#define SOCK_ACK 0x01
#define SOCK_FIN 0x10
#define SOCK_RTT_MULTIPLIER 0.125
#define SOCK_INITRTT 3000000
#define SOCK_DELACK 1
#define SOCK_NODELACK 2
#define SOCK_UNAVAIL                            0x01 /*slot in cache is used */
#define SOCK_AVAIL                              0x10 /* usused */

typedef int SockSeq;

typedef struct SockSendHdr {	/* sent over network */
     SockSeq seq;		/* seq number of packet */
     SockSeq ack;		/* base cumulative ack */
     size_t  size;		/* size of pakcet payload */
     /* 0's and 1's showing holes and recd pkts starting from 'ack' */
     u_int   ackBits;
     u_char  pktFlags;		/* is the ack valid? */
     char    padding[3];
} SockSendHdr;

typedef struct SockPktHdr {	/* locally maintained */
     SockSeq seq;		/* seq number of packet */
     SockSeq ack;		/* base cumulative ack */
     size_t  size;		/* size of pakcet payload */
     /* 0's and 1's showing holes and recd pkts starting from 'ack' */
     short   numRxmit;		/* number of retransmissions so far */
     timev   sndTime;		/* for timestamps -- unused so far */
} SockPktHdr;

typedef struct SockPkt {
     SockSendHdr sendHdr;	/* pkt header sent over network */
     SockPktHdr localHdr;	/* locally processed header fields */
     char *buf;			/* actual data */
     u_char validFlag;		/* valid entry or not (in cache)? */
     char padding[3];		/* for word alignment */
} SockPkt;

typedef struct SockCache {
     u_short fstate;		/* active, alive, closed,... */
     SockSeq lastSent;		/* seq number of last packet sent */
     SockSeq lastRead;		/* last pkt read by application */
     SockSeq firstUnack;	/* first unack'd packet */
     SockSeq lastAck;		/* seq number of last ack */
     SockSeq ackBase;		/* cumulative ack base */
     u_int   ackMask;		/* bit pattern of recd pkts and "holes" */
     int     srtt;		/* smoothed round-trip time estimate */
     int     delack;		/* number of pkts before delayed ack */
     SockPkt *pkts[SOCK_MAXWIND]; /* cache of unack'd packets */
     SockPkt *rcvPkts[SOCK_MAXWIND]; /* buf recd pkts for inorder delivery */
} SockCache;

#define SOCK_MIN(x,y) ((x)<(y)?(x):(y))

/*
#define	SOCK_SEQ_LT(a,b)	((int)((a)-(b)) < 0)
#define	SOCK_SEQ_LEQ(a,b)	((int)((a)-(b)) <= 0)
#define	SOCK_SEQ_GT(a,b)	((int)((a)-(b)) > 0)
#define	SOCK_SEQ_GEQ(a,b)	((int)((a)-(b)) >= 0)
*/

/** a simple reliability protocol built on top of UDP 
 primarily for experimentation */
class UdpReliableSocket : public UdpSocket {
public:
    UdpReliableSocket();
    ~UdpReliableSocket();
    UdpReliableSocket(int isCtrl);

    Sock_Status InitSocket();
    /** connect socket to remote socket */
    Sock_Status Connect(char *hostname, Port port);
    /** accept a connection on a port 
      return val: socket connected to hostname */
    UdpReliableSocket* Accept();

    /** recv will recv at most length bytes */
    Sock_Status Recv(char *buffer, int length);
    /** recvfull will recv exactly length bytes */
    Sock_Status RecvFull(char *buffer, int length);
    /** send length bytes from buffer */
    Sock_Status Send(char *buffer, int length);

    /** set the frequency of acknowledgements */
    inline void SetDelack(int value) {delack_ = value;};

protected:
    Sock_Status RecvReliableDgram(char *buffer, unsigned int length, 
				  timev *timeout);
    int ReadPkt(char *buffer, int length, timev *timeout);
    int SendPkt(SockPkt *pkt);
    int SendAck(int delack);
    int RexmtPkt(SockPkt *pkt);
    void FreePkts(int index, SockSeq pktSeq, SockSeq baseAck);
    void ComputeRtt(SockSeq baseAck);
    int ProcessAck(SockSeq baseAck, u_int ackMask);
    SockPkt *MakePkt(char *buffer, SockSeq seq, SockSeq ack, 
			     u_int ackBits, size_t length);
    int InsertPkt(SockPkt *pkt);

    /** cache of pkts for udp_reliable protocol */
    SockCache *cache_;		
    int delack_;		/* number of pkts before delayed ack */
};

#endif
